import os
import click
import pkg_resources
from spacectl.conf.global_conf import *
from spacectl.lib.output import print_data

__all__ = ['cli']


@click.group()
def cli():
    pass


@cli.command()
def version():
    """Print the client version information"""
    _version = _get_version_from_pkg() or _get_version_from_file() or 'unknown'
    print_data({'version': _version}, 'yaml')


def _get_version_from_pkg():
    try:
        return pkg_resources.get_distribution('spacectl').version
    except Exception:
        return None


def _get_version_from_file():
    try:
        with open(os.path.join(SRC_DIR, 'VERSION'), 'r') as f:
            return f.read().strip()

    except Exception:
        return None



import click
import abc
import os
import traceback
import shutil
from functools import wraps
from spaceone.core import utils
from spacectl.lib.apply import store


def execute_wrapper(func):
    # the instance which calls this func is bounded as self
    @wraps(func)
    def wrapper(self):
        if not self.silent:
            click.echo(f'{"  "*self.depth}TASK [{self.name}] {"*"*1000}'[:self.terminal_width])

        try:
            func(self)
        except Exception as e:
            if not self.silent:
                click.echo(f'(ERROR) =>\n{e}\n'
                           f'>> Spec: {utils.dump_json(self.spec)}',  err=True)
                click.echo('')

            _debug = os.environ.get('SPACECTL_DEBUG', 'false')

            if _debug.lower() == 'true':
                click.echo(traceback.format_exc(), err=True)

            store.increase_failure()
            exit(1)

        if not self.silent:
            if self.debug:
                if self.output_path:
                    output_data = self.output.get(self.output_path)
                else:
                    output_data = utils.dump_json(self.output, 4)

                click.echo(f'(DEBUG) =>\n{output_data}')
                click.echo('')

        store.increase_success()

        if self.is_loop:
            store.append_task_result(self.to_dict())
        else:
            store.set_task_result(self.to_dict())

    return wrapper


class BaseTask(metaclass=abc.ABCMeta):
    fields_to_apply_template = ["name", "uses", "spec", "apply_if"]

    output_path = None

    def __init__(self, task_info, silent=False, depth=0, is_loop=False):
        self.name = task_info['name']
        self.id = task_info.get('id')
        self.debug = task_info.get('debug', False)
        self.silent = silent
        self.depth = depth
        self.is_loop = is_loop
        self.state = 'IN_PROGRESS'
        self.terminal_width = shutil.get_terminal_size(fallback=(120, 50)).columns

        self.spec = {}
        self.output = {}

        self.set_spec(task_info.get('spec', {}))

    def to_dict(self):
        return {
            'name': self.name,
            'id': self.id,
            'state': self.state,
            'spec': self.spec,
            'output': self.output
        }

    def set_spec(self, spec):
        self.spec = spec

    @abc.abstractmethod
    @execute_wrapper
    def execute(self):
        pass



from spaceone.core import utils
from spacectl.lib.parser import BaseParser


class DefaultParser(BaseParser):

    def load_template(self, template):
        keys = []
        for rule in template.get('list', []):
            key, name = get_key_and_name(rule, self.use_name_alias)
            key = self._check_index_and_condition(key)
            keys.append(key)
            self._sort_map[name] = key

        self._keys = list(set(keys))

    def parse_data(self, raw_data):
        parsed_data = {}
        for rule in self.template.get('list', []):
            key, name = get_key_and_name(rule, self.use_name_alias)

            if key.startswith('tags.'):
                parsed_data[name] = get_tags_value(raw_data, key)
            else:
                parsed_data[name] = utils.get_dict_value(raw_data, key)

        return parsed_data

    @staticmethod
    def _check_index_and_condition(key):
        index_keys = key.split('.')

        i = 0
        for key in index_keys:
            try:
                if len(key) > 1 and key[0] == "?":
                    break
                else:
                    int(key)
                    break
            except Exception as e:
                pass

            i += 1

        return '.'.join(index_keys[:i])


def get_key_and_name(rule, use_name_alias=True):
    if '|' in rule:
        key, name = rule.split('|')

        if not use_name_alias:
            name = key.rsplit('.', 1)[-1]

    else:
        key = rule
        name = rule

    return key.strip(), name.strip()


def get_tags_value(data, key):
    sub_key = key[5:]
    tags = data.get('tags', {})
    return tags.get(sub_key)


if __name__ == '__main__':
    pass



from typing import Any, List
import os
import click
import jinja2
from pathlib import Path
import shutil

from spaceone.core import utils
from spacectl.lib.apply import store
from spacectl.lib.output import print_data
from spacectl.modules import MODULES


def to_data(value):
    json_value = utils.dump_json(value).strip()
    return f'<DATA>{json_value}</DATA>'


JINJA_ENV = jinja2.Environment(loader=jinja2.BaseLoader(), variable_start_string='${{', variable_end_string='}}')
JINJA_ENV.filters['to_data'] = to_data
TASK_KEYS = [
    'id',
    'name',
    'uses',
    'spec',
    'var',
    'if',
    'loop',
    'sub_tasks',
    'debug'
]


class TaskManager:

    task_queue: list = None

    def __init__(self, silent: bool, output: str = 'none'):
        self.task_queue = []
        self.silent = silent
        self.output = output
        self.terminal_width = shutil.get_terminal_size(fallback=(120, 50)).columns

    def load(self, file_path: str):
        data: dict = utils.load_yaml_from_file(file_path)

        for import_file in data.get('import', []):
            # import file path is relative to current file_path
            absolute_location = Path(file_path).parent
            self.load(os.path.join(absolute_location, import_file))

        store.set_var(data.get('var', {}))
        store.set_env(data.get('env', {}))

        for task in data.get('tasks', []):
            self._check_task(task)
            self.task_queue.append(task)

    def run(self):
        self._run_tasks(self.task_queue)
        self._finish_tasks()

    def _check_task(self, task_info: dict):
        task_json = f'>> Task: {utils.dump_json(task_info, 4)}'

        # When the type of task_info is not a dict
        if not isinstance(task_info, dict):
            raise ValueError(f'Task format is invalid.\n{task_json}')

        # Unsupported keys in task_info
        wrong_keys = list(set(task_info.keys()) - set(TASK_KEYS))
        if len(wrong_keys) > 0:
            raise ValueError(f'Unknown keys: {wrong_keys}\n{task_json}')

        uses = task_info.get('uses')
        sub_tasks = task_info.get('sub_tasks')

        # At least 1 key required (uses or sub_tasks)
        if not (uses or sub_tasks):
            raise ValueError(f'Required keys: uses or sub_tasks\n{task_json}')

        if sub_tasks:
            # When the type of sub_tasks is not a list
            if not isinstance(sub_tasks, list):
                raise ValueError(f'Type error: sub_tasks(list)\n{task_json}')

            # Recursive check of sub_tasks
            for sub_task in sub_tasks:
                self._check_task(sub_task)

    def _run_tasks(self, tasks: List[dict], depth: int = 0):
        for task_info in tasks:

            # Set task name
            task_name = task_info.get('name') or task_info.get('id') or utils.generate_id('task')
            task_info['name'] = task_name

            if 'loop' in task_info:
                loop_cond = self._parse_loop_condition(task_info['loop'], task_name)

                if not self.silent:
                    self._print_stage(task_name, 'LOOP', depth)

                for item in loop_cond:
                    self._execute_task(task_info, task_name, depth+1, item=item)

            else:
                self._execute_task(task_info, task_name, depth)

    def _execute_task(self, task_info: dict, task_name: str, depth: int, item: Any = None):
        self._update_var(task_info, task_name, item)

        task_name: str = self._render_template(task_name, item)
        if_cond: bool = self._parse_if_condition(task_info.get('if', True), task_name, item)
        is_loop: bool = True if item else False

        if if_cond:
            if 'sub_tasks' in task_info:
                if not self.silent:
                    self._print_stage(task_name, 'SUB-TASKS', depth)

                self._run_tasks(task_info['sub_tasks'], depth + 1)
            else:
                task_module = self._get_module_from_uses(task_info['uses'], task_name)
                task_info = self._change_template_values(task_info, task_name, item)

                t = task_module(task_info, silent=self.silent, depth=depth, is_loop=is_loop)
                t.execute()
        else:
            self._print_stage(task_name, 'SKIP', depth)
            store.increase_skip()

    def _update_var(self, task_info: dict, task_name: str, item: Any = None):
        if 'var' in task_info:
            try:
                data = self._render_template(task_info['var'], item)

            except Exception as e:
                raise ValueError(f'[{task_name}] Template Render Error: {e}\n'
                                 f'>> Task: {utils.dump_json(task_info, 4)}')

            store.set_var(data)

    def _get_module_from_uses(self, uses: str, task_name: str):
        try:
            uses = uses.strip()
            location, module = uses.split('/')
        except Exception as e:
            raise ValueError(f'[{task_name}] Module Load Error: uses = {uses}')

        if location == '@modules':
            self._check_default_module(module, task_name)
            return MODULES[module]
        else:
            self._check_remote_module(location, module)
            raise ValueError(f'[{task_name}] Module Load Error: external modules are not currently supported.')

    def _check_remote_module(self, location: str, module: str):
        pass

    @staticmethod
    def _check_default_module(module: str, task_name: str):
        if module not in MODULES:
            raise ValueError(f'[{task_name}] Not found module: uses = {module}')

    def _change_template_values(self, task_info: dict, task_name: str, item=None) -> dict:
        try:
            task_info = self._render_template(task_info, item)

        except Exception as e:
            raise ValueError(f'[{task_name}] Template Render Error: {e}\n'
                             f'>> Task: {utils.dump_json(task_info, 4)}')

        return task_info

    @staticmethod
    def _render_template(value: Any, item: Any = None) -> Any:
        kwargs = {
            'var': store.get_var(),
            'env': store.get_env(),
            'tasks': store.get_task_results()
        }

        if item:
            kwargs['item'] = item

        task_json: str = utils.dump_json(value)
        jinja_template = JINJA_ENV.from_string(task_json)

        template_applied_value: str = jinja_template.render(**kwargs)

        template_applied_value = template_applied_value.replace('"<DATA>', '')
        template_applied_value = template_applied_value.replace('</DATA>"', '')

        return utils.load_json(template_applied_value)

        # task_yaml: str = utils.dump_yaml(value)
        # jinja_template = JINJA_ENV.from_string(task_yaml)
        # template_applied_value: str = jinja_template.render(**kwargs)
        #
        # template_applied_value = template_applied_value.replace('None', 'null')
        # template_applied_value = template_applied_value.replace("\\\'", '"')
        #
        # return utils.load_yaml(template_applied_value)

    def _parse_if_condition(self, if_cond_str: str, task_name: str, item: Any) -> bool:
        try:
            if_cond_str = self._render_template(if_cond_str, item)
            if_cond = eval(str(if_cond_str))

        except Exception as e:
            click.echo(f'[{task_name}] If Condition Error: {if_cond_str}\n(Error) => {e}\n')
            if_cond = False

        if not isinstance(if_cond, bool):
            raise ValueError(f'[{task_name}] If Condition Error: {if_cond_str}\n'
                             f'(Type) => {type(if_cond)}\n'
                             f'(Value) => {if_cond}')

        return if_cond

    def _parse_loop_condition(self, loop_cond_str: str, task_name: str) -> list:
        try:
            loop_cond = self._render_template(loop_cond_str)

        except Exception as e:
            click.echo(f'[{task_name}] Loop Condition Error: {loop_cond_str}\n(Error) => {e}\n')
            loop_cond = []

        if loop_cond and not isinstance(loop_cond, list):
            raise ValueError(f'[{task_name}] Loop Condition Error: {loop_cond_str}\n'
                             f'(Type) => {type(loop_cond)}\n'
                             f'(Value) => {loop_cond}')

        return loop_cond or []

    def _finish_tasks(self):
        task_output = store.get_output()

        if not self.silent:
            click.echo(f'FINISHED [ ok={task_output["success"]}, skipped={task_output["skip"]} ] '
                       f'{"*" * 1000}'[:self.terminal_width])

        if not self.output == 'none':
            print_data(task_output, self.output)

        click.echo('')

    def _print_stage(self, name, action, depth):
        click.echo(f'{"  " * depth}{action} [{name}] {"*" * 1000}'[:self.terminal_width])



import io
import boto3
import pandas as pd

from spaceone.core import utils
from spacectl.lib.apply.task import BaseTask
from spacectl.lib.apply.task import execute_wrapper


class Task(BaseTask):

    def __init__(self, task_info, *args, **kwargs):
        super().__init__(task_info, *args, **kwargs)
        self.s3 = None
        self._validate()
        self._create_session()

    @execute_wrapper
    def execute(self):
        output = self.spec.get('output', 'csv')

        with io.StringIO() as io_buffer:
            if output == 'csv':
                df = pd.DataFrame(self.spec['data'])
                df.to_csv(io_buffer, index=False)
            else:
                io_buffer.write(utils.dump_json({'data': self.spec['data']}, 4))

            response = self.s3.put_object(
                Bucket=self.spec['bucket'], Key=self.spec['path'], Body=io_buffer.getvalue()
            )

            status = response.get('ResponseMetadata', {}).get('HTTPStatusCode')

            if status != 200:
                raise Exception(f'The put_object command in S3 failed. (status = {status})')

    def _create_session(self):
        if 'role_arn' in self.spec:
            session = boto3.Session(aws_access_key_id=self.spec['aws_access_key_id'],
                                    aws_secret_access_key=self.spec['aws_secret_access_key'])

            sts = session.client('sts')
            sts.get_caller_identity()

            assume_role_request = {
                'RoleArn': self.spec['role_arn'],
                'RoleSessionName': utils.generate_id('AssumeRoleSession')
            }

            if 'external_id' in self.spec:
                assume_role_request['ExternalId'] = self.spec['external_id']

            assume_role_object = sts.assume_role(**assume_role_request)
            credentials = assume_role_object['Credentials']
            self.s3 = boto3.client('s3', aws_access_key_id=credentials['AccessKeyId'],
                                   aws_secret_access_key=credentials['SecretAccessKey'],
                                   aws_session_token=credentials['SessionToken'])

        else:
            self.s3 = boto3.client('s3', aws_access_key_id=self.spec['aws_access_key_id'],
                                   aws_secret_access_key=self.spec['aws_secret_access_key'])

    def _validate(self):
        if 'aws_access_key_id' not in self.spec:
            raise ValueError(f'Required key: aws_access_key_id\n'
                             f'spec: {utils.dump_json(self.spec, 4)}')

        if 'aws_secret_access_key' not in self.spec:
            raise ValueError(f'Required key: aws_secret_access_key\n'
                             f'spec: {utils.dump_json(self.spec, 4)}')

        if 'bucket' not in self.spec:
            raise ValueError(f'Required key: bucket\n'
                             f'spec: {utils.dump_json(self.spec, 4)}')



import time
from datetime import datetime
from dateutil.relativedelta import relativedelta
from spaceone.core import utils
from spacectl.lib.apply.task import BaseTask
from spacectl.lib.apply.task import execute_wrapper


class Task(BaseTask):

    @execute_wrapper
    def execute(self):
        base_date = self.spec.get('base_date')
        
        if base_date:
            dt = utils.parse_timediff_query(base_date)
        else:
            dt = datetime.utcnow()

        self.output = {
            'year': dt.strftime('%Y'),
            'year-1y': self._change_time(dt, '%Y', years=-1),
            'year-2y': self._change_time(dt, '%Y', years=-2),
            'year-3y': self._change_time(dt, '%Y', years=-3),
            'month': dt.strftime('%Y-%m'),
            'month+1m': self._change_time(dt, '%Y-%m', months=+1),
            'month-1m': self._change_time(dt, '%Y-%m', months=-1),
            'month-2m': self._change_time(dt, '%Y-%m', months=-2),
            'month-3m': self._change_time(dt, '%Y-%m', months=-3),
            'month-4m': self._change_time(dt, '%Y-%m', months=-4),
            'month-5m': self._change_time(dt, '%Y-%m', months=-5),
            'month-6m': self._change_time(dt, '%Y-%m', months=-6),
            'month-7m': self._change_time(dt, '%Y-%m', months=-7),
            'month-8m': self._change_time(dt, '%Y-%m', months=-8),
            'month-9m': self._change_time(dt, '%Y-%m', months=-9),
            'month-10m': self._change_time(dt, '%Y-%m', months=-10),
            'month-11m': self._change_time(dt, '%Y-%m', months=-11),
            'date': dt.strftime('%Y-%m-%d'),
            'date+1d': self._change_time(dt, '%Y-%m-%d', days=+1),
            'date-1d': self._change_time(dt, '%Y-%m-%d', days=-1),
            'date-2d': self._change_time(dt, '%Y-%m-%d', days=-2),
            'date-3d': self._change_time(dt, '%Y-%m-%d', days=-3),
            'date-4d': self._change_time(dt, '%Y-%m-%d', days=-4),
            'date-5d': self._change_time(dt, '%Y-%m-%d', days=-5),
            'date-6d': self._change_time(dt, '%Y-%m-%d', days=-6),
            'date-7d': self._change_time(dt, '%Y-%m-%d', days=-7),
            'date-8d': self._change_time(dt, '%Y-%m-%d', days=-8),
            'date-9d': self._change_time(dt, '%Y-%m-%d', days=-9),
            'date-10d': self._change_time(dt, '%Y-%m-%d', days=-10),
            'date-11d': self._change_time(dt, '%Y-%m-%d', days=-11),
            'date-12d': self._change_time(dt, '%Y-%m-%d', days=-12),
            'date-13d': self._change_time(dt, '%Y-%m-%d', days=-13),
            'date-14d': self._change_time(dt, '%Y-%m-%d', days=-14),
            'date-15d': self._change_time(dt, '%Y-%m-%d', days=-15),
            'year_start_month': dt.replace(month=1, day=1).strftime('%Y-%m'),
            'month_start_date': dt.replace(day=1).strftime('%Y-%m-%d'),
            'datetime': dt.strftime('%Y-%m-%dT%H:%M:%SZ'),
            'timestamp': int(time.time())
        }

    @staticmethod
    def _change_time(dt: datetime, fmt, **kwargs):
        dt = dt + relativedelta(**kwargs)
        return dt.strftime(fmt)




from spacectl.modules import shell, print, resource, export_google_sheets, date, join, concat, export_s3, export_mysql

MODULES = {
    'shell': shell.Task,
    'print': print.Task,
    'resource': resource.Task,
    'export-google-sheets': export_google_sheets.Task,
    'export-s3': export_s3.Task,
    'export-mysql': export_mysql.Task,
    'date': date.Task,
    'join': join.Task,
    'concat': concat.Task
}


import os
from pathlib import Path

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
SRC_DIR = os.path.dirname(BASE_DIR)
DEFAULT_TEMPLATE_DIR = os.path.join(BASE_DIR, 'template')
ASSET_DIR = os.path.join(BASE_DIR, 'asset')

HOME_DIR = str(Path.home())
CONFIG_DIR = os.path.join(HOME_DIR, '.spaceone')
TEMPLATE_DIR = os.path.join(CONFIG_DIR, 'template')
ENVIRONMENT_CONF_PATH = os.path.join(CONFIG_DIR, 'environment.yml')
ENVIRONMENT_DIR = os.path.join(CONFIG_DIR, 'environments')

DEFAULT_ENVIRONMENT = 'default'
DEFAULT_PARSER = 'spacectl.lib.parser.default.DefaultParser'

RESOURCE_ALIAS = {
    # Identity
    'domain': ['identity', 'Domain'],
    'domain_owner': ['identity', 'DomainOwner'],
    'project': ['identity', 'Project'],
    'project_group': ['identity', 'ProjectGroup'],
    'pg': ['identity', 'ProjectGroup'],
    'user': ['identity', 'User'],
    'api_key': ['identity', 'APIKey'],
    'policy': ['identity', 'Policy'],
    'role': ['identity', 'Role'],
    'provider': ['identity', 'Provider'],
    'service_account': ['identity', 'ServiceAccount'],
    'sa': ['identity', 'ServiceAccount'],
    'endpoint': ['identity', 'Endpoint'],

    # Inventory
    'region': ['inventory', 'Region'],
    'server': ['inventory', 'Server'],
    'sv': ['inventory', 'Server'],
    'cloud_service_type': ['inventory', 'CloudServiceType'],
    'cloud_service': ['inventory', 'CloudService'],
    'cs': ['inventory', 'CloudService'],
    'resource_group': ['inventory', 'ResourceGroup'],
    'rg': ['inventory', 'ResourceGroup'],
    'collector': ['inventory', 'Collector'],


    # Config
    'user_config': ['config', 'UserConfig'],
    'domain_config': ['config', 'UserConfig'],

    # Repository
    'repository': ['repository', 'Repository'],
    'repo': ['repository', 'Repository'],
    'plugin': ['repository', 'Plugin'],
    'schema': ['repository', 'Schema'],

    # Secret
    'secret': ['secret', 'Secret'],

    # Monitoring
    'metric': ['monitoring', 'Metric'],
    'log': ['monitoring', 'Log'],
    'project_alert_config': ['monitoring', 'ProjectAlertConfig'],
    'escalation_policy': ['monitoring', 'EscalationPolicy'],
    'webhook': ['monitoring', 'Webhook'],
    'maintenance_window': ['monitoring', 'MaintenanceWindow'],
    'alert': ['monitoring', 'Alert'],
    'event': ['monitoring', 'Event'],

    # Notification
    'protocol': ['notification', 'Protocol'],
    'project_channel': ['notification', 'ProjectChannel'],
    'ph': ['notification', 'ProjectChannel'],
    'user_channel': ['notification', 'UserChannel'],
    'uh': ['notification', 'UserChannel'],
    'notification': ['notification', 'Notification'],
    'noti': ['notification', 'Notification'],

    # Cost Analysis
    'cost': ['cost_analysis', 'Cost'],
    'budget': ['cost_analysis', 'Budget'],
    'budget_usage': ['cost_analysis', 'BudgetUsage'],
}

EXCLUDE_APIS = [
    # 'identity.Domain.create',
    # 'identity.Domain.update',
]



import click
import copy
import os
from spaceone.core.auth.jwt import JWTUtil

from spacectl.command.execute import _check_api_permissions, _get_client,_call_api
from spacectl.conf.my_conf import get_config
from spacectl.lib.apply.task import BaseTask
from spacectl.lib.apply.task import execute_wrapper
from spacectl.modules.resource import validate
from spacectl.lib.output import echo
from spacectl.lib.template import *


class Task(BaseTask):

    def set_spec(self, spec_dict):
        self.spec = spec_dict
        if not self.spec.get("matches"):
            self.spec["matches"] = []

        self._set_mode()
        self._set_verb()

        self.spec["no_verification"] = self.spec.get("no_verification") \
            if self.spec.get("no_verification") else True

    def _set_mode(self):
        self.spec["mode"] = self.spec.get("mode", "DEFAULT")

    def _set_verb(self):
        mode = self.spec["mode"]
        custom_verb = {}
        if "verb" in self.spec:
            for k, v in self.spec["verb"].items():
                custom_verb[k] = v
        default_verb = {
            "read": "list",
            "create": "create",
            "update": "update"
        }
        validate.check_valid_verb(self.name, mode, custom_verb)
        self.spec["verb"] = default_verb
        self.spec["verb"].update(custom_verb)

    @execute_wrapper
    def execute(self):
        # execute the SpaceONE API
        service, resource = self.spec["resource_type"].split(".")
        parser = None
        output = self.spec.get('output', {})

        if output:
            options = output.get('options', {})
            add_only = options.get('add_only', True)
            template = output.get('template')

            if template == 'file':
                parser = self._load_parser(template_path=options.get('file', ''))
            elif template == 'metadata':
                parser = self._load_parser_from_metadata(options.get('metadata', ''),
                                                         options.get('use_name_alias', True))
            elif template == 'input':
                parser = self._load_parser(columns=options.get('columns', []))

            if add_only:
                self._add_only_query(parser)

        if self.spec['mode'] == 'EXEC':
            self._exec(service, resource)

        if self.spec['mode'] in ['DEFAULT', 'READ_ONLY', 'NO_UPDATE']:
            self._read(service, resource)

        if len(self.output) == 0 and self.spec['mode'] in ['DEFAULT', 'NO_UPDATE']:
            self._create(service, resource)

        if len(self.output) == 1 and self.spec['mode'] == 'DEFAULT':
            self._update(service, resource)

        if parser:
            if isinstance(self.output, list):
                self.output = [parser.parse_data(res) for res in self.output]
            elif isinstance(self.output, dict):
                self.output = parser.parse_data(self.output)

    def _exec(self, service, resource):
        try:
            verb = self.spec["verb"]["exec"]
            # echo("Start " + ".".join([service, resource, verb]), flag=not self.silent)
            _check_api_permissions(service, resource, verb)

            pagination = self.spec.get('pagination')
            self.output = _execute_api(service, resource, verb, self.spec.get("data", {}), api_version="v1",
                                       output="yaml", parser={}, silent=self.silent, pagination=pagination)
            # echo("Finish " + ".".join([service, resource, verb]), flag=not self.silent)
            # echo(f'### {verb} Response ###', flag=not self.silent)
            # echo(self.output, flag=not self.silent)
        except Exception as e:
            raise e

    def _read(self, service, resource):
        try:
            read_params = {match: self.spec["data"][match] for match in self.spec["matches"]}
            verb = self.spec["verb"]["read"]

            read_resources = _execute_api(service, resource, verb, read_params, api_version="v1", output="yaml", parser={}, silent=self.silent)
            if isinstance(read_resources, list):
                length = len(read_resources)
                if length == 0:
                    self.output = {}
                elif length >= 1:
                    self.output = read_resources[0]

                if length >= 2:
                    click.echo("Multiple resources are searched so select the first one.")

            elif isinstance(read_resources, dict):  # like dict
                self.output = read_resources
            echo(f'### {verb} Response ###', flag=not self.silent)
            echo(read_resources, flag=not self.silent)
        except Exception as e:
            click.echo(e, err=True)

    def _update(self, service, resource):
        try:
            verb = self.spec["verb"]["update"]
            echo("Start " + ".".join([service, resource, verb]), flag=not self.silent)
            _check_api_permissions(service, resource, verb)
            self.output = _execute_api(service, resource, verb, self.spec['data'], api_version="v1", output="yaml", parser={}, silent=self.silent)
            echo("Finish " + ".".join([service, resource, verb]), flag=not self.silent)
            echo(f'### {verb} Response ###', flag=not self.silent)
            echo(self.output, flag=not self.silent)
        except Exception as e:
            echo(e, flag=True, err=True)
            echo(
                "Unavailable update field so Skip" +
                ".".join([service, resource, "update"]),
                flag=True, err=True, terminate=True
            )

    def _create(self, service, resource):
        verb = self.spec["verb"]["create"]
        echo("Start " + ".".join([service, resource, verb]), flag=not self.silent)
        create_result = _execute_api(service, resource, verb, self.spec['data'], api_version="v1", output="yaml", parser={}, silent=self.silent)
        echo("Finished " + ".".join([service, resource, verb]), flag=not self.silent)
        self.output = create_result
        echo(f'### {verb} Response ###', flag=not self.silent)
        echo(self.output, flag=not self.silent)

    def __str__(self):
        result_dict = {}
        for field in self.fields_to_apply_template:
            result_dict[field] = getattr(self, field)
        return str(field)

    @staticmethod
    def _load_parser(columns=None, template_path=None):
        template = load_template(None, None, columns, template_path=template_path)
        return load_parser(None, None, template)

    def _load_parser_from_metadata(self, metadata, use_name_alias):
        cloud_service_type = self._get_cloud_service_type(metadata)
        user_config = self._get_user_config(metadata)
        show_optional = False

        if user_config:
            table_fields = user_config.get('data', {}).get('options', {}).get('fields', [])
            show_optional = True
        else:
            table_fields = cloud_service_type.get('metadata', {}).get('view', {}).get('table', {}).get('layout', {}).get(
                'options', {}).get('fields', [])

        template = self._generate_template_from_metadata_table_fields(table_fields, show_optional)

        if not user_config:
            template['template']['list'] = ['project_id', 'provider|Provider',
                                            'region_code|Region', 'name|Name'] \
                                           + template['template']['list'] \
                                           + ['reference.resource_id|Resource ID']

        return load_parser(None, None, template, use_name_alias)

    def _add_only_query(self, parser):
        query = self.spec['data'].get('query', {})
        query['only'] = parser.keys
        self.spec['data']['query'] = query

    def _generate_template_from_metadata_table_fields(self, table_fields, show_optional):
        _list = []
        for _field in table_fields:
            if show_optional:
                _list.append(self._set_field(_field))
            else:
                if not _field.get('options', {}).get('is_optional', False):
                    # Temporary Code
                    if _field.get('key') not in ['project_id', 'provider', 'region_code', 'name', 'updated_at',
                                                 'launched_at']:
                        _list.append(self._set_field(_field))

        return {'template': {'list': _list}}

    @staticmethod
    def _set_field(field):
        key = field.get('key')
        if key:
            _f = key

            sub_key = field.get('options', {}).get('sub_key')
            if sub_key:
                _f = f'{_f}.{sub_key}'

            name = field.get('name')
            if name:
                return f'{_f}|{name}'
            else:
                return _f

    def _get_resource_type_from_metadata(self, metadata):
        try:
            provider, cloud_service_group, cloud_service_type = metadata.split('.')
            return provider, cloud_service_group, cloud_service_type
        except Exception as e:
            raise Exception(e)

    def _get_user_config_name(self, metadata, user):
        provider, cloud_service_group, cloud_service_type = self._get_resource_type_from_metadata(metadata)
        return f"console:USER:{user}:page-schema:inventory.CloudService?provider={provider}&cloud_service_group={cloud_service_group}&cloud_service_type={cloud_service_type}:table"

    def _get_user_name(self):
        api_key = os.environ.get('SPACECTL_API_KEY', get_config().get('api_key'))
        user_name = JWTUtil.unverified_decode(api_key).get('aud', '')
        return user_name

    def _get_user_config(self, metadata):
        verb = 'list'
        params = {
            'name': self._get_user_config_name(metadata, self._get_user_name())
        }
        responses = _execute_api('config', 'UserConfig', verb, params, api_version="v1", output="json", parser={}, silent=True)

        if responses:
            return responses[0]
        else:
            return {}

    def _get_cloud_service_type(self, metadata):
        provider, cloud_service_group, cloud_service_type = self._get_resource_type_from_metadata(metadata)

        verb = 'list'
        params = {
            'provider': provider,
            'group': cloud_service_group,
            'name': cloud_service_type
        }
        responses = _execute_api('inventory', 'CloudServiceType', verb, params,
                                 api_version="v1", output="json", parser={}, silent=True)

        if len(responses) > 0:
            return responses[0]
        else:
            return {}


def _execute_api(service, resource, verb, params={}, api_version='v1', output='yaml', parser=None,
                 silent=False, pagination=None):
    config = get_config()
    _check_api_permissions(service, resource, verb)
    client = _get_client(service, api_version)

    # _call_api can change some data of params so need deepcopy
    # e.g. credential of identity.Token

    if pagination:
        page_size = pagination.get('size', 1000)
        req_params = copy.deepcopy(params)
        response = {
            'results': [],
            'total_count': 0
        }
        req_params['query'] = req_params.get('query', {})
        req_params['query']['count_only'] = True
        count_response_stream = _call_api(client, resource, verb, req_params, config=config)
        for count_response in count_response_stream:
            response['total_count'] = count_response.get('total_count', 0)

        req_params['query']['count_only'] = False
        page_count = int(response['total_count'] / page_size) + 1

        click.echo(f'Pagination ({service}.{resource}.{verb}): total_count={response["total_count"]}, '
                   f'page_size={page_size}, page_count={page_count}')

        for page_num in range(page_count):
            req_params['query']['page'] = {
                'start': (page_size * page_num) + 1,
                'limit': page_size
            }
            p_response_stream = _call_api(client, resource, verb, req_params, config=config)
            for p_response in p_response_stream:
                response['results'] += p_response.get('results', [])

        response_stream = [response]
    else:
        response_stream = _call_api(client, resource, verb, copy.deepcopy(params), config=config)

    for response in response_stream:
        if verb in ['list', 'stat', 'analyze']:
            results = response.get('results', [])

            if len(results) == 0:
                return []
            elif len(results) > 0:
                return results
            else:
                Exception()
        elif verb == 'create':
            return response
        elif verb == 'update':
            return response
        else:
            return response



from spaceone.core import utils
from spacectl.lib.apply.task import BaseTask
from spacectl.lib.apply.task import execute_wrapper
from spacectl.lib.apply.store import get_task_result


class Task(BaseTask):

    @execute_wrapper
    def execute(self):
        self._validate()

        origin_task_result = get_task_result(self.id) or {}
        origin_output = origin_task_result.get('output', [])

        self.output = origin_output + self.spec.get('data', [])

    def _validate(self):
        if 'data' not in self.spec:
            raise ValueError(f'Required key: data\n'
                             f'spec: {utils.dump_json(self.spec, 4)}')



import click
from spacectl.lib.output import print_data
from spacectl.conf.global_conf import DEFAULT_ENVIRONMENT
from spacectl.conf.my_conf import *

__all__ = ['cli']


@click.group()
def cli():
    pass


@cli.group()
def config():
    """Modify spaceconfig files"""
    pass


@config.command()
@click.option('-e', '--environment', prompt='Environment', help='Environment', default=DEFAULT_ENVIRONMENT)
@click.option('-f', '--import-file', type=click.Path(exists=True), help='YAML file only')
def init(environment, import_file):
    """Initialize spaceconfig"""
    set_environment(environment)

    if import_file:
        import_config(import_file, environment)
    else:
        set_config({}, environment)


@config.command()
@click.argument('key')
@click.argument('value')
def set(key, value):
    """Set specific spaceconfig"""
    data = get_config()
    data[key] = value
    set_config(data)


@config.command()
@click.argument('key')
def remove(key):
    """Remove specific spaceconfig"""
    data = get_config()

    if key in data:
        del data[key]

    set_config(data)


@config.command()
@click.option('-s', '--switch', help='Switch the environment')
@click.option('-r', '--remove', help='Remove the environment')
def environment(switch, remove):
    """Manage environments"""

    if switch:
        environments = list_environments()
        if switch not in environments:
            raise Exception(f"'{switch}' environment not found.")

        set_environment(switch)
        click.echo(f"Switched to '{switch}' environment.")
    elif remove:
        environments = list_environments()
        if remove not in environments:
            raise Exception(f"'{switch}' environment not found.")

        remove_environment(remove)
        click.echo(f"'{remove}' environment has been removed.")
    else:
        try:
            current_env = get_environment()
        except Exception:
            current_env = None

        environments = list_environments()
        if len(environments) == 0:
            raise Exception('spaceconfig is undefined. (Use "spacectl config init")')

        for env in environments:
            if current_env == env:
                click.echo(f'{env} (current)')
            else:
                click.echo(env)


@config.command()
@click.option('-o', '--output', default='yaml', help='Output format',
              type=click.Choice(['json', 'yaml']), show_default=True)
def show(output):
    """Display a spaceconfig"""
    data = get_config()
    print_data(data, output)


@config.group()
def endpoint():
    """Manage API endpoints"""
    pass


@endpoint.command()
@click.argument('service')
@click.argument('endpoint')
def add(service, endpoint):
    """Add a API endpoint"""
    try:
        endpoints = get_endpoint()
    except Exception:
        endpoints = {}

    if service in endpoints:
        raise ValueError(f"'{service}' service already exists.")
    endpoints[service] = endpoint
    set_endpoint(endpoints)
    click.echo(f"'{service}: {endpoint}' endpoint has been added.")


@endpoint.command()
@click.argument('service')
def remove(service):
    """Remove a API endpoint"""
    remove_endpoint(service)
    click.echo(f"'{service}' endpoint has been removed.")


@endpoint.command()
@click.option('-o', '--output', default='table', help='Output format',
              type=click.Choice(['table', 'json', 'yaml']), show_default=True)
def show(output):
    """Display API endpoints"""
    endpoints = list_endpoints()
    print_data(endpoints, output, headers=['Service', 'Endpoint'])



import os
import click
from spacectl.lib.apply.task_manager import TaskManager
from spacectl.lib.apply import store


@click.group()
def cli():
    pass


@cli.command()
@click.argument('file_path')
@click.option('-o', '--output', default='none', help='Output format',
              type=click.Choice(['json', 'yaml', 'none']), show_default=True)
@click.option('-e', '--env', 'env', multiple=True, help='Configure additional environment variables.')
@click.option('-v', '--var', 'var', multiple=True,
              help='Configure additional variables. You can override existing variables as well.')
@click.option('--var-file', type=click.Path(exists=True), help='Yaml file to configure var and env')
@click.option('-s', '--silent', is_flag=True, default=False, show_default=True)
def apply(file_path, output, env, var, var_file, silent):
    store.init_env(env)
    store.init_var(var_file, var)
    task_manager = TaskManager(silent, output)
    task_manager.load(file_path)

    task_manager.run()


if __name__ == '__main__':
    apply()



import click
from spaceone.core import utils
from spacectl.lib.apply.task import BaseTask
from spacectl.lib.apply.task import execute_wrapper
from spacectl.lib.output import print_data


class Task(BaseTask):

    output_path = 'result'

    @execute_wrapper
    def execute(self):
        self._validate()

        output = self.spec.get('output', 'json')
        print_data(self.spec['data'], output)
        click.echo('')

    def _validate(self):
        if 'data' not in self.spec:
            raise ValueError(f'Required key: data\n'
                             f'spec: {utils.dump_json(self.spec, 4)}')



import os
from spaceone.core import utils

_DATA = {
    'var': {},
    'env': {},
    'tasks': {},
    'results': [],
    'success': 0,
    'skip': 0,
    'failure': 0,
}


def init_env(env: tuple):
    _DATA['env'] = dict(os.environ)
    set_env(_parse_inputs(env))


def init_var(var_file: str, var: tuple):
    if var_file:
        data = utils.load_yaml_from_file(var_file)

        if isinstance(data, dict):
            set_var(data.get('var', {}))

    set_var(_parse_inputs(var))


def get_var(key: str = None, default=None):
    if key:
        return _DATA['var'].get(key, default)
    else:
        return _DATA['var']


def set_var(data: dict):
    _DATA['var'].update(data)


def get_env(key: str = None, default=None):
    if key:
        return _DATA['env'].get(key, default)
    else:
        return _DATA['env']


def set_env(data: dict):
    _DATA['env'].update(data)


def get_task_results():
    return _DATA['tasks']


def get_task_result(task_id: str):
    return _DATA['tasks'].get(task_id)


def append_task_result(task_result: dict):
    task_id = task_result.get('id')
    if task_id:
        if task_id not in _DATA['tasks']:
            _DATA['tasks'][task_id] = []

        _DATA['tasks'][task_id].append(task_result)

    # _DATA['results'].append(task_result)


def set_task_result(task_result: dict):
    task_id = task_result.get('id')
    if task_id:
        _DATA['tasks'][task_id] = task_result

    # _DATA['results'].append(task_result)


def increase_success():
    _DATA['success'] += 1


def increase_failure():
    _DATA['failure'] += 1


def increase_skip():
    _DATA['skip'] += 1


def get_output():
    return {
        'var': _DATA['var'],
        'results': _DATA['results'],
        'success': _DATA['success'],
        'skip': _DATA['skip']
    }


def _parse_inputs(inputs: tuple):
    result = {}
    for data in inputs:
        i_split = data.split('=')
        if len(i_split) == 2:
            result[i_split[0]] = i_split[1]
        else:
            raise ValueError(f'Input parameter({data}) is invalid. (format: key=value)')

    return result



from abc import *


class BaseParser(metaclass=ABCMeta):

    def __init__(self, template, use_name_alias=True):
        self._template = template
        self._keys = []
        self._sort_map = {}
        self.use_name_alias = use_name_alias
        self.load_template(template)

    @property
    def template(self):
        return self._template

    @property
    def keys(self):
        return self._keys

    @abstractmethod
    def load_template(self, template):
        pass

    @abstractmethod
    def parse_data(self, raw_data):
        pass

    def get_sort_key(self, key):
        return self._sort_map.get(key, key)



import pymysql

from spaceone.core import utils
from spacectl.lib.apply.task import BaseTask
from spacectl.lib.apply.task import execute_wrapper


class Task(BaseTask):

    def __init__(self, task_info, *args, **kwargs):
        super().__init__(task_info, *args, **kwargs)
        self.conn = None
        self._validate()
        self._create_session()

    @execute_wrapper
    def execute(self):
        data = self.spec.get('data', [])
        for item in data:
            self._insert_data(item)

    def _convert_value(self, value):
        if isinstance(value, str):
            return f'"{value}"'
        else:
            return str(value)

    def _insert_data(self, item):
        keys_str = ", ".join(item.keys())
        values_str = ", ".join(map(self._convert_value, item.values()))
        sql = f'INSERT INTO {self.spec["table"]} ({keys_str}) VALUES ({values_str})'

        self.conn.cursor().execute(sql)
        self.conn.commit()

    def _create_session(self):
        conn_info = {
            'host': self.spec['host'],
            'port': self.spec.get('port', 3306),
            'user': self.spec['user'],
            'password': self.spec['password'],
            'db': self.spec['db'],
            'charset': 'utf8'
        }

        if 'pem' in self.spec:
            conn_info['ssl_ca'] = self.spec['pem']

        self.conn = pymysql.connect(**conn_info)

    def _validate(self):
        if 'host' not in self.spec:
            raise ValueError(f'Required key: host\n'
                             f'spec: {utils.dump_json(self.spec, 4)}')

        if 'user' not in self.spec:
            raise ValueError(f'Required key: user\n'
                             f'spec: {utils.dump_json(self.spec, 4)}')

        if 'password' not in self.spec:
            raise ValueError(f'Required key: password\n'
                             f'spec: {utils.dump_json(self.spec, 4)}')

        if 'db' not in self.spec:
            raise ValueError(f'Required key: db\n'
                             f'spec: {utils.dump_json(self.spec, 4)}')

        if 'table' not in self.spec:
            raise ValueError(f'Required key: table\n'
                             f'spec: {utils.dump_json(self.spec, 4)}')



import os
from spaceone.core import utils
from spacectl.conf.global_conf import *

__all__ = ['set_environment', 'get_environment', 'remove_environment', 'list_environments',
           'import_config', 'set_config', 'get_config', 'set_endpoint', 'get_endpoint',
           'remove_endpoint', 'list_endpoints', 'set_template', 'get_template', 'remove_template']


def set_environment(environment):
    utils.create_dir(CONFIG_DIR)
    utils.create_dir(ENVIRONMENT_DIR)
    utils.save_yaml_to_file({'environment': environment}, ENVIRONMENT_CONF_PATH)


def get_environment():
    try:
        data = utils.load_yaml_from_file(ENVIRONMENT_CONF_PATH)
    except Exception:
        raise Exception('spaceconfig is undefined. (Use "spacectl config init")')

    environment = data.get('environment')
    if not environment:
        raise Exception('The environment is not set. Switch the environment. (Use "spacectl config environment --help")')

    return environment


def remove_environment(environment):
    try:
        environment_path = os.path.join(ENVIRONMENT_DIR, f'{environment}.yml')
        if os.path.exists(environment_path):
            os.remove(environment_path)
    except Exception as e:
        raise Exception(f'Environment deletion error: {e}')

    environments = list_environments()
    if len(environments) > 0:
        utils.save_yaml_to_file({'environment': environments[0]}, ENVIRONMENT_CONF_PATH)
    else:
        os.remove(ENVIRONMENT_CONF_PATH)


def list_environments():
    environments = []
    try:
        for f in os.listdir(ENVIRONMENT_DIR):
            if os.path.isfile(os.path.join(ENVIRONMENT_DIR, f)) and f.find('.yml') > 1:
                environments.append(f.rsplit('.', 1)[0])
    except Exception:
        raise Exception('spaceconfig is undefined. (Use "spacectl config init")')

    return environments


def import_config(import_file_path, environment=None):
    if environment is None:
        environment = get_environment()

    try:
        environment_path = os.path.join(ENVIRONMENT_DIR, f'{environment}.yml')
        data = utils.load_yaml_from_file(import_file_path)
        utils.save_yaml_to_file(data, environment_path)
    except Exception:
        raise Exception(f'Import file format is invalid. (file = {import_file_path})')


def set_config(new_data, environment=None):
    if environment is None:
        environment = get_environment()

    try:
        environment_path = os.path.join(ENVIRONMENT_DIR, f'{environment}.yml')
        utils.save_yaml_to_file(new_data, environment_path)
    except Exception:
        raise Exception('spaceconfig is undefined. (Use "spacectl config init")')


def get_config(key=None, default=None, environment=None):
    if environment is None:
        environment = get_environment()

    try:
        environment_path = os.path.join(ENVIRONMENT_DIR, f'{environment}.yml')
        data = utils.load_yaml_from_file(environment_path)
    except Exception:
        raise Exception('spaceconfig is undefined. (Use "spacectl config init")')

    if key:
        return data.get(key, default)
    else:
        return data


def set_endpoint(endpoints, environment=None):
    data = get_config(environment)
    data['endpoints'] = endpoints

    set_config(data, environment)


def get_endpoint(service=None, environment=None):
    endpoints = get_config('endpoints', {}, environment)

    if service:
        return endpoints.get(service)
    else:
        return endpoints


def remove_endpoint(service, environment=None):
    data = get_config(environment)
    endpoints = data.get('endpoints', {})

    if service in endpoints:
        del endpoints[service]

    data['endpoints'] = endpoints

    set_config(data, environment)


def list_endpoints(environment=None):
    endpoints = get_config('endpoints', {}, environment)
    result = []

    for service, endpoint in endpoints.items():
        result.append((service, endpoint))

    return result


def set_template(service, resource, data):
    utils.create_dir(TEMPLATE_DIR)
    my_template_path = os.path.join(TEMPLATE_DIR, f'{service}.{resource}.yml')
    utils.save_yaml_to_file(data, my_template_path)


def remove_template(service, resource):
    my_template_path = os.path.join(TEMPLATE_DIR, f'{service}.{resource}.yml')
    if os.path.exists(my_template_path):
        os.remove(my_template_path)


def get_template(service, resource):
    return _get_my_template(service, resource) or _get_default_template(service, resource)


def _get_my_template(service, resource):
    try:
        my_template_path = os.path.join(TEMPLATE_DIR, f'{service}.{resource}.yml')
        data = utils.load_yaml_from_file(my_template_path)
        data['type'] = 'custom'
        return data
    except Exception:
        return None


def _get_default_template(service, resource):
    try:
        default_template_path = os.path.join(DEFAULT_TEMPLATE_DIR, f'{service}.{resource}.yml')
        data = utils.load_yaml_from_file(default_template_path)
        data['type'] = 'default'
        return data
    except Exception:
        return None



import click


def check_valid_verb(task_name, mode, custom_verb):
    if mode == 'DEFAULT':
        error_if_invalid_verb(task_name, mode, ['exec'], custom_verb)
    if mode == "READ_ONLY":
        error_if_invalid_verb(task_name, mode, ['create', 'update', 'exec'], custom_verb)
    if mode == "NO_UPDATE":
        error_if_invalid_verb(task_name, mode, ['update', 'exec'], custom_verb)
    if mode == 'EXEC':
        error_if_invalid_verb(task_name, mode, ['read', 'create', 'update'], custom_verb)


def error_if_invalid_verb(task_name, mode, verb_types, custom_verb):
    for verb_type in verb_types:
        if verb_type in custom_verb:
            click.echo("You cannot define {verb_name} as {verb_type} with {mode} in task-{task_name}".format(
                verb_name=custom_verb[verb_type],
                verb_type=verb_type,
                mode=mode,
                task_name=task_name
            ), err=True)
            exit(1)



import click
import csv
import sys
from tabulate import tabulate
from spaceone.core import utils


def print_data(data, output, **kwargs):
    if 'root_key' in kwargs:
        data = utils.get_dict_value(data, kwargs['root_key'], [])
        del kwargs['root_key']

    if output == 'table':
        if len(data) == 0:
            echo('NO DATA')
        else:
            _print_table(data, **kwargs)
    elif output == 'json':
        _print_json(data, **kwargs)
    elif output == 'yaml':
        _print_yaml(data, **kwargs)
    elif output == 'csv':
        _print_csv(data, **kwargs)
    elif output == 'quiet':
        _print_quiet(data, **kwargs)
    elif output == 'text':
        echo(data)


def _print_table(data, **kwargs):
    data, headers, total_count = _parse_data_by_options(data, **kwargs)

    if isinstance(data, dict):
        _print_yaml(data)
    else:
        click.echo(tabulate(data, tablefmt='presto', headers=headers or 'keys'))

        if total_count:
            click.echo()
            click.echo(f' Count: {len(data)} / {int(total_count)}')


def _print_csv(data, **kwargs):
    data, headers, total_count = _parse_data_by_options(data, **kwargs)

    if isinstance(data, dict):
        _print_yaml(data)
    else:
        if headers:
            writer = csv.writer(sys.stdout)
            writer.writerow(headers)
            writer.writerows(data)
        else:
            writer = csv.DictWriter(sys.stdout, fieldnames=data[0].keys())
            writer.writeheader()
            writer.writerows(data)


def _parse_data_by_options(data, **kwargs):
    headers = kwargs.get('headers')
    total_count = kwargs.get('total_count')

    if isinstance(data, list) and len(data) > 0 and not isinstance(data[0], (dict, list, tuple)):
        headers = ['Values']
        data = [[v] for v in data]

    return data, headers, total_count


def _print_json(data, **kwargs):
    if data == {}:
        click.echo()
    else:
        click.echo(utils.dump_json(data, indent=4))


def _print_yaml(data, **kwargs):
    if data == {}:
        click.echo()
    else:
        click.echo('---')
        click.echo(utils.dump_yaml(data))


def _print_quiet(data, **kwargs):
    for d in data:
        items = list(d.values())
        if len(items) != 1:
            click.echo("Please Selector only one column for quiet output.", err=True)
            exit(1)

        click.echo(items[0])


def echo(message, flag=True, err=False, terminate=False):
    if flag:
        click.echo(message, err=err)
    if terminate:
        exit(1)



import subprocess
from spaceone.core import utils
from spacectl.lib.apply.task import BaseTask
from spacectl.lib.apply.task import execute_wrapper


class Task(BaseTask):

    output_path = 'result'

    @execute_wrapper
    def execute(self):
        self._validate()

        stdout = subprocess.PIPE
        stderr = subprocess.STDOUT
        completed_process = subprocess.run(
            [
                "/bin/bash",
                "-c",
                self.spec['run']
            ],
            stdout=stdout,
            stderr=stderr,
        )

        if completed_process.returncode != 0:
            raise Exception(completed_process.stdout.decode('utf-8'))

        self.output = {
            "result": completed_process.stdout.decode('utf-8'),
            "return_code": completed_process.returncode
        }

    def _validate(self):
        if 'run' not in self.spec:
            raise ValueError(f'Required key: run\n'
                             f'spec: {utils.dump_json(self.spec, 4)}')



#!/usr/bin/env python3

import os
import sys
import click
import traceback

try:
    import spacectl
except Exception:
    path = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    if path not in sys.path:
        sys.path.append(path)
from spacectl.command import apply, config, execute, version, api_resource, template

_DEBUG = os.environ.get('SPACECTL_DEBUG', 'false')
_HELP = """
spacectl controls the SpaceONE services.\n
API Reference: https://spaceone-dev.gitbook.io/spaceone-apis\n
Following steps for first time user.\n
    1. spacectl config init\n
    2. spacectl config set api_key <api_key>\n
    3. spacectl config endpoint add <service> <endpoint>
"""

cli = click.CommandCollection(sources=[apply.cli, config.cli, execute.cli, version.cli,
                                       api_resource.cli, template.cli], help=_HELP)


def main():
    try:
        cli(prog_name='spacectl')
    except Exception as e:
        click.echo(f'ERROR: {e}')
        click.echo()

        if _DEBUG.lower() == 'true':
            click.echo(traceback.format_exc())


if __name__ == '__main__':
    main()



import click

from spaceone.core.error import ERROR_BASE
from spaceone.core import pygrpc
from spaceone.core.utils import parse_endpoint

from spacectl.lib.output import print_data
from spacectl.conf.global_conf import RESOURCE_ALIAS
from spacectl.conf.my_conf import get_config, get_endpoint

__all__ = ['cli']


@click.group()
def cli():
    pass


@cli.command()
def api_resources():
    """Print the supported API resources"""
    environment = get_config('environment')
    endpoints = get_endpoint(environment)
    resources = _get_resources_from_client(endpoints)
    resources = _get_resource_alias(resources)
    _print_api_resources(resources)


def _get_resources_from_client(endpoints, api_version='v1'):
    resources = {}
    for service, endpoint in endpoints.items():
        try:
            e = parse_endpoint(endpoint)
            protocol = e.get('scheme')

            if protocol not in ['grpc', 'grpc+ssl']:
                raise ValueError(f'Unsupported protocol. (supported_protocol=grpc|grpc+ssl, endpoint={endpoint})')

            if protocol == 'grpc+ssl':
                ssl_enabled = True
            else:
                ssl_enabled = False

            client = pygrpc.client(endpoint=f'{e.get("hostname")}:{e.get("port")}',
                                   version=api_version, ssl_enabled=ssl_enabled)

            resources[service] = {}
            for api_resource, verb in client.api_resources.items():
                resources[service][api_resource] = {
                    'alias': [],
                    'verb': verb
                }
        except ERROR_BASE as e:
            raise e
        except ValueError as e:
            raise e
        except Exception:
            raise ValueError(f'Endpoint is invalid. (endpoint={endpoint})')

    return resources


def _get_resource_alias(resources):
    for alias, resource_type in RESOURCE_ALIAS.items():
        service, resource = resource_type
        if service in resources:
            if resource in resources[service]:
                resources[service][resource]['alias'].append(alias)

    return resources


def _print_api_resources(resources):
    data = []
    for service, service_value in resources.items():
        for resource, resource_value in service_value.items():
            data.append((service, resource, ', '.join(resource_value['alias']), ', '.join(resource_value['verb'])))

    print_data(data, 'table', headers=['Service', 'Resource', 'Short Names', 'Verb'])



import click
import fnmatch
import types
import os

from google.protobuf.json_format import MessageToDict
from spaceone.core.error import ERROR_BASE
from spaceone.core import pygrpc
from spaceone.core.utils import parse_endpoint, load_json

from spacectl.lib.output import print_data
from spacectl.lib.template import *
from spacectl.conf.global_conf import RESOURCE_ALIAS, EXCLUDE_APIS
from spacectl.conf.my_conf import get_config, get_endpoint

__all__ = ['cli']


@click.group()
def cli():
    pass


@cli.command()
@click.argument('resource')
@click.option('-p', '--parameter', multiple=True, help='Input Parameter (-p <key>=<value> -p ...)')
@click.option('-j', '--json-parameter', help='JSON parameter')
@click.option('-f', '--file-parameter', 'file_path', type=click.Path(exists=True), help='YAML file only')
@click.option('-v', '--api-version', default='v1', help='API Version', show_default=True)
@click.option('-o', '--output', default='yaml', help='Output format',
              type=click.Choice(['json', 'yaml']), show_default=True)
def get(resource, parameter, json_parameter, file_path, api_version, output):
    """Show details of a specific resource"""
    service, resource = _get_service_and_resource(resource)
    params = _parse_parameter(file_path, json_parameter, parameter)
    _execute_api(service, resource, 'get', 'get', params=params, api_version=api_version, output=output)


@cli.command()
@click.argument('resource')
@click.option('-p', '--parameter', multiple=True, help='Parameter (-p <key>=<value>)')
@click.option('-j', '--json-parameter', help='JSON type parameter')
@click.option('-f', '--file-parameter', 'file_path', type=click.Path(exists=True), help='Parameter file (YAML)')
@click.option('-m', '--minimal', 'minimal_columns', is_flag=True, help='Minimal columns')
@click.option('-a', '--all', 'all_columns', is_flag=True, help='All columns')
@click.option('-c', '--columns', help='Specific columns (-c id,name)')
@click.option('-t', '--template-file', 'template_path', type=click.Path(exists=True), help='Template file (YAML)')
@click.option('-l', '--limit', type=int, help='Number of rows')
@click.option('-s', '--sort', help="Sorting by given key (-s [-]<key>)")
@click.option('-m', '--method', default='list', help='List API method', show_default=True)
@click.option('-v', '--api-version', default='v1', help='API Version', show_default=True)
@click.option('-o', '--output', default='table', help='Output format',
              type=click.Choice(['table', 'json', 'yaml', 'csv', 'quiet']), show_default=True)
def list(resource, parameter, json_parameter, file_path, minimal_columns, all_columns, columns, template_path,
         limit, sort, method, api_version, output):
    """Display one or many resources"""
    service, resource = _get_service_and_resource(resource)

    if columns:
        columns = columns.split(',')

    template = load_template(service, resource, columns, template_path=template_path)
    parser = None

    params = _parse_parameter(file_path, json_parameter, parameter)
    params['query'] = params.get('query', {})

    if all_columns:
        params['query']['minimal'] = False
    elif minimal_columns:
        params['query']['minimal'] = True
    elif template:
        parser = load_parser(service, resource, template)
        params['query']['only'] = parser.keys

    if limit:
        params['query']['page'] = {'limit': limit}

    if sort:
        if sort.startswith('-'):
            desc = True
            sort_key = sort[1:]
        else:
            desc = False
            sort_key = sort

        if parser:
            sort_key = parser.get_sort_key(sort_key) or sort_key

        params['query']['sort'] = {
            'key': sort_key,
            'desc': desc
        }

    _execute_api(service, resource, method, 'list', params=params, api_version=api_version, output=output, parser=parser)


@cli.command()
@click.argument('resource')
@click.option('-p', '--parameter', multiple=True, help='Input Parameter (-p <key>=<value> -p ...)')
@click.option('-j', '--json-parameter', help='JSON type parameter')
@click.option('-f', '--file-parameter', 'file_path', type=click.Path(exists=True), help='YAML file only')
@click.option('-c', '--columns', help='Specific columns (-c id,name)')
@click.option('-l', '--limit', type=int, help='Number of rows')
@click.option('-m', '--method', default='stat', help='Stat API method', show_default=True)
@click.option('-v', '--api-version', default='v1', help='API Version', show_default=True)
@click.option('-o', '--output', default='table', help='Output format',
              type=click.Choice(['table', 'json', 'yaml', 'csv', 'quiet']), show_default=True)
def stat(resource, parameter, json_parameter, file_path, columns, limit, method, api_version, output):
    """Querying statistics for resources"""
    service, resource = _get_service_and_resource(resource)
    parser = None

    if columns:
        columns = columns.split(',')
        template = load_template(service, resource, columns)
        parser = load_parser(service, resource, template)

    params = _parse_parameter(file_path, json_parameter, parameter)

    if limit:
        if service == 'statistics' and resource == 'Resource':
            params['page'] = {'limit': limit}
        else:
            params['query'] = params.get('query', {})
            params['query']['page'] = {'limit': limit}

    _execute_api(service, resource, method, 'stat', params=params, api_version=api_version, output=output, parser=parser)


@cli.command()
@click.argument('verb')
@click.argument('resource')
@click.option('-p', '--parameter', multiple=True, help='Input Parameter (-p <key>=<value> -p ...)')
@click.option('-j', '--json-parameter', help='JSON parameter')
@click.option('-f', '--file-parameter', 'file_path', type=click.Path(exists=True), help='YAML file only')
@click.option('-v', '--api-version', default='v1', help='API Version', show_default=True)
@click.option('-o', '--output', default='yaml', help='Output format',
              type=click.Choice(['table', 'csv', 'json', 'yaml']), show_default=True)
def exec(verb, resource, parameter, json_parameter, file_path, api_version, output):
    """Execute a method to resource"""
    service, resource = _get_service_and_resource(resource)
    params = _parse_parameter(file_path, json_parameter, parameter)
    _execute_api(service, resource, verb, 'exec', params=params, api_version=api_version, output=output)


def _parse_parameter(file_parameter=None, json_parameter=None, parameter=None):
    if parameter is None:
        parameter = []

    if file_parameter:
        params = load_yaml_from_file(file_parameter)
    else:
        params = {}

    if json_parameter:
        json_params = load_json(json_parameter)
        params.update(json_params)

    for p in parameter:
        p_split = p.split('=')
        if len(p_split) == 2:
            params[p_split[0]] = p_split[1]
        else:
            raise ValueError(f'Input parameter({p}) is invalid. (format: key=value)')

    return params


def _execute_api(service, resource, verb, command, params=None, api_version='v1', output='yaml', parser=None):
    if params is None:
        params = {}

    config = get_config()
    _check_api_permissions(service, resource, verb)
    client = _get_client(service, api_version)
    response_stream = _call_api(client, resource, verb, params, config=config)

    for response in response_stream:
        if command in ['list', 'stat'] and parser:
            results = []
            try:
                for result in response.get('results', []):
                    results.append(parser.parse_data(result))
            except Exception:
                raise Exception(f'{service}.{resource} template format is invalid.')

            response['results'] = results

        options = {}

        if 'total_count' in response:
            options['total_count'] = response['total_count']

        if output in ['table', 'csv', 'quiet'] and 'results' in response:
            options['root_key'] = 'results'

            print_data(response, output, **options)
        else:
            print_data(response, output)


def _check_api_permissions(service, resource, verb):
    request_api = f'{service}.{resource}.{verb}'

    for exclude_api in EXCLUDE_APIS:
        if fnmatch.fnmatch(request_api, exclude_api):
            raise Exception(f"{request_api} not allowed.")


def _get_service_and_resource(resource):
    if resource in RESOURCE_ALIAS:
        return RESOURCE_ALIAS[resource]
    else:
        resource_split = resource.split('.')
        if len(resource_split) != 2:
            raise ValueError(f'Resource format is invalid. (resource = <service>.<resource>)')
        return resource_split[0], resource_split[1]


def _get_client(service, api_version):
    endpoint = get_endpoint(service)

    if endpoint is None:
        raise Exception(f'Endpoint is not set. (service={service})')

    try:
        e = parse_endpoint(endpoint)

        protocol = e.get('scheme')

        if protocol not in ['grpc', 'grpc+ssl']:
            raise ValueError(f'Unsupported protocol. (supported protocol = grpc | grpc+ssl, endpoint={endpoint})')

        if protocol == 'grpc+ssl':
            ssl_enabled = True
        else:
            ssl_enabled = False

        client = pygrpc.client(endpoint=f'{e.get("hostname")}:{e.get("port")}',
                               version=api_version, ssl_enabled=ssl_enabled, max_message_length=1024*1024*256)
        client.service = service
        client.api_version = api_version
        return client
    except ERROR_BASE as e:
        raise e
    except Exception:
        raise ValueError(f'Endpoint is invalid. (endpoint={endpoint})')


def _check_resource_and_verb(client, resource, verb):
    if not hasattr(client, resource):
        raise Exception(f"'{client.service}.{client.api_version}' service does not have a '{resource}' resource.")

    if not hasattr(getattr(client, resource), verb):
        raise Exception(f"'{client.service}.{client.api_version}.{resource}' resource does not have a '{verb}' verb.")


def _call_api(client, resource, verb, params=None, **kwargs):
    if params is None:
        params = {}

    _check_resource_and_verb(client, resource, verb)

    config = kwargs.get('config', {})
    api_key = os.environ.get('SPACECTL_API_KEY', config.get('api_key'))
    domain_id = os.environ.get('SPACECTL_DOMAIN_ID', config.get('domain_id'))

    if domain_id:
        params['domain_id'] = config.get('domain_id')

    try:
        metadata = (()) if api_key is None else (('token', api_key),)
        resource_client = getattr(client, resource)
        resource_verb = getattr(resource_client, verb)
        response_or_iterator = resource_verb(
            params,
            metadata=metadata
        )
        if isinstance(response_or_iterator, Exception):
            raise response_or_iterator
        elif isinstance(response_or_iterator, types.GeneratorType):
            for response in response_or_iterator:
                yield _change_message(response)
        else:
            yield _change_message(response_or_iterator)
    except ERROR_BASE as e:
        if e.error_code == 'ERROR_AUTHENTICATE_FAILURE':
            raise Exception('The api_key of spaceconfig is not set. (Use "spacectl config init")')
        else:
            raise Exception(e.message.strip())
    except Exception as e:
        if hasattr(e, 'exception'):
            raise Exception(e.exception())
        else:
            raise Exception(e)


def _change_message(message):
    return MessageToDict(message, preserving_proto_field_name=True)



import click
from spaceone.core.utils import load_yaml_from_file
from spacectl.lib.output import print_data
from spacectl.conf.global_conf import RESOURCE_ALIAS, DEFAULT_PARSER
from spacectl.conf.my_conf import set_template, remove_template, get_template

__all__ = ['cli']


@click.group()
def cli():
    pass


@cli.group()
def template():
    """Manage resource templates"""
    pass


@template.command()
@click.argument('resource')
@click.option('-f', '--file', 'file_path', type=click.Path(exists=True), help='Import template file (YAML)')
def set(resource, file_path):
    """Set resource template"""
    service, resource = _get_service_and_resource(resource)

    if file_path:
        template = load_yaml_from_file(file_path)
        set_template(service, resource, template)

    else:
        raise Exception("'--file' option is required.")


@template.command()
@click.argument('resource')
def remove(resource):
    """Remove resource template"""
    service, resource = _get_service_and_resource(resource)
    remove_template(service, resource)


@template.command()
@click.argument('resource')
def show(resource):
    """Display resource template"""
    service, resource = _get_service_and_resource(resource)
    template = get_template(service, resource)

    if template:
        template['parser'] = template.get('parser', DEFAULT_PARSER)
        print_data(template, 'yaml')
    else:
        click.echo(f'{service}.{resource} template is undefined.')
        click.echo()


def _get_service_and_resource(resource):
    if resource in RESOURCE_ALIAS:
        return RESOURCE_ALIAS[resource]
    else:
        resource_split = resource.split('.')
        if len(resource_split) != 2:
            raise ValueError(f'Resource format is invalid. (resource = <service>.<resource>)')
        return resource_split[0], resource_split[1]


#
#   Copyright 2020 The SpaceONE Authors.
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.


from setuptools import setup, find_packages

with open('VERSION', 'r') as f:
    VERSION = f.read().strip()
    f.close()

setup(
    name='wizctl',
    version=VERSION,
    description='CloudWiz Command Line Interface',
    long_description='',
    url='https://cloudwiz.ktds.com/',
    author='Cloudwiz Team',
    author_email='admin@cloudwiz.ktds.com',
    license='Apache License 2.0',
    packages=find_packages(),
    install_requires=[
        'kt-ds-core',
        'kt-ds-api',
        'Click',
        'tabulate',
        'jinja2',
        'gspread',
        'pandas',
        'google-api-core',
        'google-api-python-client',
        'oauth2client',
        'boto3',
        'PyMySQL'
    ],
    package_data={
        'spacectl': ['template/*.yml']
    },
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'wizctl = spacectl.main:main',
        ]
    },
)



from spaceone.core.utils import load_yaml_from_file

from spacectl.conf.global_conf import DEFAULT_PARSER
from spacectl.conf.my_conf import get_template


def load_template(service, resource, columns, template_path=None):
    if columns:
        template = {
            'template': {
                'list': columns
            }
        }
    else:
        if template_path:
            template = load_yaml_from_file(template_path)
        else:
            template = get_template(service, resource)
    return template


def load_parser(service, resource, template, use_name_alias=True):
    parser = template.get('parser', DEFAULT_PARSER)
    template = template.get('template')

    if parser is None:
        raise Exception(f"'parser' is undefined in {service}.{resource} template.")

    if template is None:
        raise Exception(f"'template' is undefined in {service}.{resource} template.")

    try:
        module_name, class_name = parser.rsplit('.', 1)
        parser_module = __import__(module_name, fromlist=[class_name])
    except Exception:
        raise Exception(f'Parser is invalid. ({parser})')

    try:
        return getattr(parser_module, class_name)(template, use_name_alias)
    except Exception:
        raise Exception(f'{service}.{resource} template format is invalid.')


import pandas as pd
import numpy as np
from spaceone.core import utils
from spacectl.lib.apply.task import BaseTask
from spacectl.lib.apply.task import execute_wrapper
from spacectl.lib.template import *


class Task(BaseTask):

    @execute_wrapper
    def execute(self):
        self._validate()

        join_keys = self.spec.get('join_keys')
        how = self.spec.get('how', 'left')
        columns = self.spec.get('output', {}).get('columns')
        append_origin_columns = self.spec.get('output', {}).get('append_origin_columns', {})
        origin_columns = []
        origin_columns_location = append_origin_columns.get('location')
        origin_columns_start = append_origin_columns.get('start', 0)
        origin_columns_end = append_origin_columns.get('end')

        df = pd.DataFrame(self.spec.get('origin_data', []))

        if origin_columns_location:
            origin_columns = list(df.columns)
            if origin_columns_end:
                origin_columns = origin_columns[origin_columns_start:origin_columns_end]
            else:
                origin_columns = origin_columns[origin_columns_start:]

        if len(df) > 0:
            other_df = pd.DataFrame(self.spec.get('other_data', []))

            if len(other_df) > 0:
                df = df.merge(other_df, on=join_keys, how=how)
                df = df.replace({np.nan: None})

        results = df.to_dict('records')

        if columns:
            if origin_columns_location == 'LEFT':
                columns = origin_columns + columns
            elif origin_columns_location == 'RIGHT':
                columns = columns + origin_columns

            parser = self._load_parser(columns)
            results = [parser.parse_data(res) for res in results]

        self.output = results

    @staticmethod
    def _load_parser(columns):
        template = load_template(None, None, columns)
        return load_parser(None, None, template)

    def _validate(self):
        if 'origin_data' not in self.spec:
            raise ValueError(f'Required key: origin_data\n'
                             f'spec: {utils.dump_json(self.spec, 4)}')

        if 'other_data' not in self.spec:
            raise ValueError(f'Required key: other_data\n'
                             f'spec: {utils.dump_json(self.spec, 4)}')

        if 'join_keys' not in self.spec:
            raise ValueError(f'Required key: join_keys\n'
                             f'spec: {utils.dump_json(self.spec, 4)}')



import gspread
import re
import string
import pandas as pd
import time
from googleapiclient.discovery import build
from oauth2client.service_account import ServiceAccountCredentials

from spaceone.core import utils
from spacectl.lib.apply.task import BaseTask
from spacectl.lib.apply.task import execute_wrapper
from spacectl.lib.output import echo

DEFAULT_HEADER_CELL = 'A1'
SCOPES = [
    'https://www.googleapis.com/auth/drive',
    'https://www.googleapis.com/auth/spreadsheets'
]
# DEFAULT_QUERY = f"'{ROOT_FOLDER_ID}' in parents and trashed = false"
FILE_QUERY = "mimeType != 'application/vnd.google-apps.folder'"
FOLDER_QUERY = "mimeType = 'application/vnd.google-apps.folder'"


class Task(BaseTask):

    def __init__(self, task_info, *args, **kwargs):
        super().__init__(task_info, *args, **kwargs)
        self._validate()

        service_account_json = self.spec.get('service_account_json')
        creds = ServiceAccountCredentials.from_json_keyfile_name(service_account_json, scopes=SCOPES)
        self.gc = gspread.authorize(creds)
        self.drive_svc = build('drive', 'v3', credentials=creds)
        self.sheet_svc = build('sheets', 'v4', credentials=creds)
        self.root_folder_id = self.spec.get('folder_id')

    @execute_wrapper
    def execute(self):
        path = self.spec.get('path')

        folders, sheet_name = self._get_folder_and_sheet(path)
        folder_id = self._get_folder_id(self.root_folder_id, folders)
        sheet_id = self._get_sheet_id(folder_id, sheet_name)
        sheet = self._get_sheets(sheet_id)

        self.clear_all_worksheet(sheet)
        self.export_data(sheet)
        time.sleep(3)

        copy = self.spec.get('copy')
        if copy:
            self._copy_sheet(copy, sheet_id)

    def _copy_sheet(self, copy_path, sheet_id):
        folders, sheet_name = self._get_folder_and_sheet(copy_path)
        folder_id = self._get_folder_id(self.root_folder_id, folders)
        self._delete_sheet(folder_id, sheet_name)

        echo(f'copy file: {sheet_id} -> {copy_path}')
        body = {
            'name': sheet_name,
            'parents': [folder_id]
        }
        self.drive_svc.files().copy(fileId=sheet_id, body=body, supportsAllDrives=True).execute()

    def _delete_sheet(self, parent_id, name):
        query = f"name = '{name}' and {FILE_QUERY} and '{parent_id}' in parents"
        results = self.drive_svc.files().list(**self._get_search_options(query)).execute()

        items = results.get('files', [])

        body = {'trashed': True}

        for item in items:
            echo(f'delete file: {item["id"]}')
            self.drive_svc.files().update(fileId=item['id'], body=body, supportsAllDrives=True).execute()

    def _get_sheet_id(self, parent_id, name):
        query = f"name = '{name}' and {FILE_QUERY} and '{parent_id}' in parents"
        echo(f'search sheet: {query}')

        results = self.drive_svc.files().list(**self._get_search_options(query)).execute()

        items = results.get('files', [])

        if len(items) > 0:
            return items[0]['id']
        else:
            return self._create_sheet(parent_id, name)

    def _create_sheet(self, parent_id, name):
        sh = self.gc.create(name)
        sheet_id = sh.id
        self._move_file(parent_id, sheet_id)

        return sheet_id

    def _move_file(self, parent_id, file_id):
        file = self.drive_svc.files().get(fileId=file_id, fields='parents').execute()
        previous_parents = ",".join(file.get('parents'))

        echo(f'move sheet: {file_id} ({previous_parents} -> {parent_id})')
        self.drive_svc.files().update(fileId=file_id, addParents=parent_id, removeParents=previous_parents,
                                      supportsAllDrives=True, fields='id, parents').execute()

    def _get_folder_id(self, parent_id, folders):
        if len(folders) == 0:
            return parent_id

        name = folders[0]
        query = f"name = '{name}' and {FOLDER_QUERY} and '{parent_id}' in parents"
        echo(f'search folder: {query}')
        results = self.drive_svc.files().list(**self._get_search_options(query)).execute()

        items = results.get('files', [])

        if len(items) > 0:
            folder_id = items[0]['id']
        else:
            folder_id = self._create_folder(parent_id, name)

        return self._get_folder_id(folder_id, folders[1:])

    def _create_folder(self, parent_id, name):
        echo(f'create folder: {name} in {parent_id}')
        metadata = {
            'name': name,
            'mimeType': 'application/vnd.google-apps.folder',
            'parents': [parent_id]
        }

        f = self.drive_svc.files().create(body=metadata, supportsAllDrives=True, fields='id').execute()
        folder_id = f.get('id')
        return folder_id

    @staticmethod
    def _get_search_options(query):
        return {
            'q': query,
            'includeItemsFromAllDrives': True,
            'supportsAllDrives': True,
            'pageSize': 10,
            'fields': 'nextPageToken, files(id, name)'
        }

    @staticmethod
    def _get_folder_and_sheet(path):
        if isinstance(path, list):
            items = path
        else:
            items = str(path).split('/')
        return items[:-1], items[-1:][0]

    def _get_sheets(self, sheet_id):
        echo(f'open sheet : {sheet_id}')
        return self.gc.open_by_key(sheet_id)

    @staticmethod
    def _set_auto_resize_columns(sheet, worksheet):
        body = {
            "requests": [
                {
                    "autoResizeDimensions": {
                        "dimensions": {
                            "sheetId": worksheet.id,
                            "dimension": "COLUMNS",
                            "startIndex": 0,
                            "endIndex": 26
                        }
                    }
                }
            ]
        }

        sheet.batch_update(body)

    def export_data(self, sheet):
        fill_na = self.spec.get('fill_na')
        for idx, raw_data in enumerate(self.spec.get('data', [])):
            if isinstance(raw_data, dict) and 'input' in raw_data:
                task = raw_data.get('input')
                options = raw_data.get('options', {})
            else:
                task = raw_data
                options = {}

            if task is None:
                echo(f'skip input: {raw_data}')
            else:
                output = task.get('output', [])

                worksheet_name = self.set_worksheet_name(task)

                if len(output) > 0:
                    worksheet = self.select_worksheet(sheet, idx, worksheet_name)

                    echo(f"export worksheet: {worksheet_name}")

                    self.export_worksheet(worksheet, output, fill_na, **options)
                    # self._set_auto_resize_columns(sheet, worksheet)
                    time.sleep(5)

                else:
                    echo(f'skip worksheet: {worksheet_name}')

    def select_worksheet(self, sheet, idx, worksheet_title):
        try:
            if self.spec.get('reset', False) is True and idx == 0:
                worksheet = sheet.get_worksheet(0)
                worksheet.update_title(worksheet_title)
            else:
                worksheet = sheet.worksheet(worksheet_title)
            return worksheet
        except Exception:
            return sheet.add_worksheet(title=worksheet_title, rows=1000, cols=26)

    def export_worksheet(self, worksheet, data, fill_na=None, **options):
        df = pd.DataFrame(data)

        if fill_na is not None:
            df = df.fillna(fill_na)

        headers = df.columns.values.tolist()
        self._format_header(worksheet, DEFAULT_HEADER_CELL, headers)
        self._format_values(worksheet, options)
        export_values = []

        for row in list(df.values):
            changed_row = []
            for value in row:
                if isinstance(value, list):
                    changed_row.append('\n'.join(value))
                else:
                    changed_row.append(value)
            export_values.append(changed_row)

        worksheet.update(DEFAULT_HEADER_CELL, [headers] + export_values)

    def clear_all_worksheet(self, sheet):
        reset = self.spec.get('reset', False)

        if reset:
            echo(f'clear all worksheet: {sheet.id}')
            sheet.add_worksheet(title='', rows=1000, cols=26)

            for worksheet in sheet.worksheets()[:-1]:
                sheet.del_worksheet(worksheet)

    @staticmethod
    def set_worksheet_name(task):
        return task.get('name', '')

    @staticmethod
    def _format_values(worksheet, options):
        decimal = options.get('decimal')

        if decimal:
            worksheet.format('2:1000', {'numberFormat': {'type': 'NUMBER', 'pattern': f'0.{ "0"*(decimal-1) }#'}})

    @staticmethod
    def _format_header(worksheet, start_at, headers):
        header_format = {
            'horizontalAlignment': 'CENTER',
            'borders': {
                'top': {'style': 'SOLID'},
                'bottom': {'style': 'SOLID'},
                'left': {'style': 'SOLID'},
                'right': {'style': 'SOLID'}
            },
            "backgroundColor": {
                "red": 12,
                "green": 12,
                "blue": 12
            },
            'textFormat': {'bold': True}
        }

        try:
            header_length = len(headers)
            _header_split = re.split('(\d+)', start_at)
            _col = _header_split[0]
            _col_idx = string.ascii_uppercase.index(_col)
            _chr_index = 64 + _col_idx + header_length if 64 + _col_idx + header_length <= 90 else 90
            _num = _header_split[1]

            header_cells = f'{start_at}:{chr(_chr_index)}{_num}'

            worksheet.format(header_cells, header_format)
        except Exception as e:
            raise Exception(f'Header Write Error => {e}')

    def _validate(self):
        if 'service_account_json' not in self.spec:
            raise ValueError(f'Required key: service_account_json\n'
                             f'spec: {utils.dump_json(self.spec, 4)}')

        if 'data' not in self.spec:
            raise ValueError(f'Required key: data\n'
                             f'spec: {utils.dump_json(self.spec, 4)}')

        if 'path' not in self.spec:
            raise ValueError(f'Required key: path\n'
                             f'spec: {utils.dump_json(self.spec, 4)}')

        if 'folder_id' not in self.spec:
            raise ValueError(f'Required key: folder_id\n'
                             f'spec: {utils.dump_json(self.spec, 4)}')